RPL

****************************************************************
* Fast Circle V2.6
* Algorithme de cration de cercle (Mid-Point)
* Exercice pour formation  l'assembleur SASM
* Yann Collet, 10/08/08
* Thanks to : Cyrille de Brebisson, for providing ideas & sample source code
****************************************************************
* Entre : Pile : 4 : Grob [Optional]  <-- If none provided, draw in GRAPH
*                 3 : Xcenter (en pixel) <= 250 000
*                 2 : Ycenter (en pixel) <= 250 000
*                 1 : Radius  (en pixel) <= 250 000
****************************************************************

::
  CK3&Dispatch

  # 111
  ::
    ROTDUP %0< IT :: % 1048576 %+ ; COERCE   ( x )
    ROTDUP %0< IT :: % 1048576 %+ ; COERCE   ( y )
    ROT COERCE                               ( Radius >= 0 )

    DEPTH THREE #> ITE 4PICK DUP             ( Vrifie Stack4 = Grob )
    DUPTYPEGROB? ?SKIP :: DROP CHECKPICT GBUFF ;  ( Sinon, fourni GBUFF )
    4UNROLL
    
**********************************************************
* Principe : x + y <= R + MaxErr
*
* Initialise : MaxErr = R (Car [(R+1)-R]/2 = R + 1/2)
*	On translate et inverse : Err = R (on teste le carry)
*              On commence avec : x=R, y=0,
* Au lieu de calculer x+y, on va juste prendre en compte les Delta
*              NextDeltaX = 2X-1
*              NextDeltaY = 2Y+1
*
* Puis on augmente rcursivement : y=y+1
*              Err = Err - NextDeltaY
*      Si Err < 0, faut aussi diminuer x=x-1
*              Err = Err + NextDeltaX
*      Puis on dessine
*      Jusqu' y>x
*
* Pour chaque point calcul, on dessine 8 pts :
* (x,y) (x,-y) (-x,y) (-x,-y) (y,x) (y,-x) (-y,x) (-y,-x)
*
**********************************************************
* MAP
* R0 : Err(A)
* R1 : x  (A)
* R2 : x0 (A)
* R3 : y  (A)
* R4 : y0 (A)
* D0 : Multiplication code
* D1 : Grob.Pointer (MaxY)
**********************************************************

CODE

  GOSBVL =SAVPTR         Sauve B.A, D.A, D0 & D1 mais dtruit C

*** Create Memory for Multiplication Code
  LC(5) #00040           C=64d soit 32 octets
  GOSBVL =MAKE$N         Creates character string in tempob area ;
* C.A = nibbles -> D0 = addr of body of $
* Note : modify A.W,B.W,C.W,D.W,R0,D0&D1
	 CD0EX
	 R0=C
	
  GOSBVL =GETPTR         Restore B.A, D.A, D0 & D1 ; modify C
	
*** Cration de R1
  GOSBVL =POP#           A.A = R ( Note : C.A & D.A modifis , D1 utilis)
  R1=A                   R1=X=R

*** Cration de R3 : y (=0)
  A=0 W                  Nettoyage pour shift plus tard
  R3=A                   R3.A= y = 0

*** Cration R4 ( y0 ) & R2 ( x0 )
  GOSBVL =POP2#          A.A = X0 ; C.A = Y0
  R2=A
  R4=C

  GOSBVL =SAVPTR         Sauve B.A, D.A, D0 & D1, mais dtruit C

*** Create Err (R0) and Restore D0
	 A=R1                   A=X=R=Err
	 AR0EX                  Err; A=D0
	 D0=A

*** PointerGrob
  A=DAT1 A               Grob
  D1=A
  D1=D1+ 15              MaxX

*** Cration Code Multiplication
  A=DAT1 A                MaxX (Note : A dj t nettoy par un A.W=0)
  A=A+CON A,7         [8] Pour passer l'octet suivant
  A=A+A A             [6]
  ASR A               [7]
  A=A+A A             [6] A.A = NibbleWidth

  C=D0
  D=C A               Sauvegarde D0
  GOSHORT START_MULT
MULT_LOOP
  DAT0=C B
  D0=D0+ 2
START_MULT
  ?A=0 A
  GOYES END_MULT
  ?A=0 P
  GOYES MULT_CSL
  ?ABIT=0 0
  GOYES MULT_CC
ADD_BC
  LC(2) #1C           B=B+C A (invers)
  ABIT=0 0            Mise  jour de A
  GOSHORT MULT_LOOP
MULT_CSL
  LC(2) #2F           CSL A (invers)
  ASR A               Mise  jour de A
  GOSHORT MULT_LOOP
MULT_CC
  LC(2) #6C           C=C+C A (invers)
  ASRB              	 Mise  jour de A (c'est pour a qu'il faut nettoyer)
  GOSHORT MULT_LOOP

END_MULT
  LC(4) #2B18         PC=A (invers)
  DAT0=C 4

  C=D A
  D0=C                Restore D0	
  D1=D1- 5            D1->MaxY

**********************************************************
* MAP
* R0 : Err(A)
* R1 : x  (A)
* R2 : x0 (A)
* R3 : y  (A)
* R4 : y0 (A)
* D0 : Multiplication code
* D1 : Grob.Pointer (MaxY)
**********************************************************
* Init pour DRAW : A=X, C=Y
  A=R1
  C=R3
BOUCLE
  GOSUB DRAW_EIGHT_POINTS    Note : A=X, C=Y
  B=C A             B=Y
  C=R0.F A
  D=C A             D=Err
  GOSUB Y_PLUS_ONE  Note : A=X; B=Y; D=Err;
  C=D A
  R0=C.F A          Save Err
  C=B A
  R3=C.F A          Save Y
  ?A>=C A           As long as X>=Y
  GOYES BOUCLE
  GOVLNG =GETPTRLOOP  Si Y>X, on arrte

Y_PLUS_ONE
* Working Register Map : A=X; B=Y; D=Err;
  C=B A             [7]
  B=B+1 A           [7] B.A = Y+1
  C=C+B A           [7] C.A = NextDeltaY
  D=D-C A           [7] D.A = NewErr
  GOC X_MINUS_ONE   (3-10)
  RTN               [9]

X_MINUS_ONE
  C=A A             [7]
  A=A-1 A           [7] A.A = X-1
  C=C+A A           [7] C.A = NextDeltaX
  D=D+C A           [7] D.A = NewErr
  R1=A.F A          Save X
  RTN

**********************************************************
* MAP
* R0 : Err(A)
* R1 : x  (A)
* R2 : x0 (A)
* R3 : y  (A)
* R4 : y0 (A)
* D0 : Multiplication code
* D1 : Grob.Pointer (MaxY)
**********************************************************

DRAW_EIGHT_POINTS
* Inherited : A=X, C=Y
  ST=1 0
DRAW_FOUR_POINTS
  B=A A                     B=x
  D=C A                     D=y
  GOSUB CALCULATE_PIXEL

  C=R1.F A
  C=-C A
  B=C A                     B=-x
  C=R3.F A
  D=C A                     D=y
  GOSUB CALCULATE_PIXEL

  C=R1.F A
  B=C A                     B=x
  C=R3.F A
  C=-C A
  D=C A                     D=-y
  GOSUB CALCULATE_PIXEL

  C=R1.F A
  C=-C A
  B=C A                     B=-x
  C=R3.F A
  C=-C A
  D=C A                     D=-y
  GOSUB CALCULATE_PIXEL

  A=R1.F A                  Swap x & y
  C=A A                     C=NewY=OldX
  AR3EX                     A=NewX=OldY
  R1=A.F A

  ?ST=0 0
  RTNYES
  ST=0 0
  GOTO DRAW_FOUR_POINTS

**********************************************************
* MAP
* R0 : Err(A)
* R1 : x  (A)
* R2 : x0 (A)
* R3 : y  (A)
* R4 : y0 (A)
* D0 : Multiplication code
* D1 : Grob.Pointer (MaxY)
**********************************************************
CALCULATE_PIXEL
* Recenter x,y using x0 & y0; Need : B=x, D=y;
  A=R2.F A            [9]  A.X = x0
  A=A+B A             [6]  A.X = Xpixel
  C=R4.F A            [9]  C.X = y0
  C=C+D A             [6]  C.X = Ypixel

CHECK_PIXEL
* Vrifie d'abord qu'on est bien dans la partie autorise de l'cran
  D=C A               [6]  D.X = y
  C=DAT1 A                 C.A = MaxY
  ?D>=C A             (9-11)
  RTNYES              Sortie si Y >= MaxY

  D1=D1+ 5
  C=DAT1 A            [18] C.A = MaxX = C.X car MaxX<=1000
  D1=D1- 5                 Retour  MaxY, pour rester standard
  ?A>=C A             (9-11)
  RTNYES              Sortie si X >= MaxX
  

DRAW_PIXEL
* On dessine un pixel  l'cran
* D1 sur GBUFF; A.X=x; B.X=NibbleWidth; C.X=(Free); D.X=y

  C=A P                    C.0 = lowX
  CDEX A                   C = y ; On sauve lowx dans D

  A=A+A A             [6]  Note A.A = x <=250000, donc pas de dbordement
  A=A+A A             [6]
  ASR A               [7]  A.A / 4
  B=A A               [7]  Nibble dans la ligne
  B=B+CON A,10        [10] Offset, car D1 est sur MaxY

  A=D0                [16] Adresse Code Multiplication
  APCEX               [16] Va  la multiplication; nibble x C.A ajout  B.A

  A=D1                [16] A=D1 c'est en fait AD1EX(8);D1=A(8)
  A=A+B A             [7]
  AD1EX               [8]  D1 pointe sur le bon nibble; A.A --> MaxY

* Allumage du bon pixel
* Technique de slection du masque utilisant P, par Detlef Muller

  LC(1) #C            [3]  A.0 = #Bh = 1100b
  C=C!D P             [5]  C.0 = 12/13/14/15 (P=0)
  P=C 0               [6]
  LC(4) #1248         [6]  C.S = Mask
  P= 0                [2]

  A=DAT1  S           [18]
  A=A!C   S           [5]  Trigger pixel
  DAT1=A  S           [17]

  D1=A                [8]  Re-init D1 sur MaxY
  
  RTN                 [9]

ENDCODE

  DROP
  ;
;


